#!/usr/local/bin/perl
#2_parse_GenBank_file.pl
#Author: Paul Stothard
#
#This script parses a GenBank record, and generates a FASTA file
#of protein translations described in the record.
#
#USAGE: perl 2_parse_GenBank_file.pl [-arguments]
#-i [FILE]     : input file (Required)
#-o [FILE]     : output file (Required)
#
#perl 2_parse_GenBank_file.pl -i NC_009673.gbk -o NC_009673.fasta

use strict;
use warnings;

use Getopt::Long;
use Data::Dumper;

#The following are part of BioPerl (http://www.bioperl.org), a
#collection of useful Perl modules for bioinformatics.
use Bio::SeqIO;
use Bio::Seq;

my %options = (
    input_file  => undef,
    output_file => undef
);

GetOptions(
    'i=s' => \$options{input_file},
    'o=s' => \$options{output_file}
);

if ( !defined( $options{input_file} ) ) {
    print_usage();
}
if ( !defined( $options{output_file} ) ) {
    print_usage();
}

#In programming terms, an object is a collection of information and
#functions (also called methods). Usually the information and
#functions are grouped into an object because they are related in some
#meaningful way. For example, an object for representing DNA sequences
#might contain the DNA sequence itself, along with functions for
#manipulating the sequence. Objects can contain other objects. For
#example, a DNA sequence object might contain a list of feature
#objects, describing features (exons, introns, coding regions, etc)
#associated with the sequence.

#Create a new Bio::SeqIO object by passing it the format of the
#sequence file ('genbank' in this case) and the name of the file
#(stored in $options{input_file}). The object that is returned is
#stored in $in. To learn more about Bio::SeqIO, see the following:
#http://doc.bioperl.org/releases/bioperl-current/bioperl-live/Bio/Seq.html
my $in = Bio::SeqIO->new(
    -format => "genbank",
    -file   => $options{input_file}
);

#Use Data::Dumper to view the information stored in $in.
print "The contents of the Bio::SeqIO object:\n";
print Dumper($in);

#Bio::SeqIO objects include a 'next_seq' function, which can be used
#to obtain a Bio::Seq object. This object represents a nucleotide
#sequence and any associated features obtained from the GenBank
#file. For more information about Bio::Seq, see the following:
#http://doc.bioperl.org/releases/bioperl-current/bioperl-live/Bio/Seq.html
my $seq = $in->next_seq();

#Use Data::Dumper to view the information stored in $seq.
print "The contents of the Bio::Seq object:\n";
print Dumper($seq);

#The Bio::Seq get_SeqFeatures() function returns an array of
#Bio::SeqFeatureI objects. These objects represent the features
#associated with the nucleotide sequence. To learn more about
#Bio::SeqFeatureI, see
#http://doc.bioperl.org/releases/bioperl-current/bioperl-live/Bio/SeqFeatureI.html
my @features = $seq->get_SeqFeatures();

#Pass @features to 'get_coding_sequence_features', as a
#reference. This subroutine, which is defined later in the script,
#returns a reference to a new array. The new array only contains
#features that represent coding sequences.
my $coding_sequence_features = get_coding_sequence_features( \@features );

#Use Data::Dumper to view the information stored in
#$coding_sequence_features.
print "The contents of \$coding_sequence_features:\n";
print Dumper($coding_sequence_features);

#Pass the array reference stored in $coding_sequence_features to
#a subroutine that returns a list of protein names and translations
#(extracted from the sequence feature objects).
my $names_and_translations
    = get_names_and_translations($coding_sequence_features);

#Use Data::Dumper to view the information stored in
#$names_and_translations.
print "The contents of \$names_and_translations:\n";
print Dumper($names_and_translations);

open( my $FILE, '>', $options{output_file} )
    or die("Cannot open file $options{output_file}: $!\n");

#The values in the $names_and_translations array are references to
#hashes (see the 'get_names_and_translations' code). The hashes
#contain two keys: 'name' and 'translation'. The 'foreach' loop below
#is used to iterate through the hash references. During each
#iteration, the current hash reference is stored in $info. To access a
#value using a key and a hash reference, use '->' between the variable
#name and the key name ($info->{name} for example).
foreach ( @{$names_and_translations} ) {
    my $info = $_;
    if ( ( defined( $info->{name} ) ) && ( defined( $info->{translation} ) ) )
    {
        print( $FILE ">$info->{name}\n$info->{translation}\n" );
    }
}

close($FILE) or die("Cannot close file: $!");

#Go through a list of Bio::SeqFeatureI objects, and return those that
#have a 'primary_tag' value of 'cds'.
sub get_coding_sequence_features {
    my $features                 = shift;
    my @coding_sequence_features = ();

    foreach ( @{$features} ) {
        my $feature = $_;

        #The 'primary_tag' method returns the feature type.
        my $type = $feature->primary_tag;
        $type = lc($type);

        #'cds' stands for coding sequence.
        if ( $type ne 'cds' ) {
            next;
        }

        #Add the feature object to @coding_sequence_features.
        push( @coding_sequence_features, $feature );
    }

    #Return a reference to @coding_sequence_features.
    return \@coding_sequence_features;
}

#Go through a list of Bio::SeqFeatureI objects, and return an array of
#hashes containing protein names and translations.
sub get_names_and_translations {
    my $features        = shift;
    my @array_of_hashes = ();

    foreach ( @{$features} ) {
        my $feature = $_;

        #Create a hash for storing a protein name and translation.
        my %info_hash = (
            name        => undef,
            translation => undef
        );

        #Use the 'has_tag' method of Bio::SeqFeatureI to determine
        #whether there is a 'product' entry associated with the
        #feature object. If there is a 'product' entry, get the values
        #associated with the entry using 'get_tag_values'. Store the
        #first value in %info_hash, under the key 'name'.
        if ( $feature->has_tag('product') ) {
            my @product_values = $feature->get_tag_values('product');
            $info_hash{name} = $product_values[0];
        }

        #Use a similar strategy to get the sequence of the protein, if
        #available.
        if ( $feature->has_tag('translation') ) {
            my @translation_values = $feature->get_tag_values('translation');
            $info_hash{translation} = $translation_values[0];
            $info_hash{translation} =~ s/[^A-Z]//ig;
        }

        #Add a reference to the recently created hash to
        #@array_of_hashes.
        push( @array_of_hashes, \%info_hash );
    }

    #Return a reference to @array_of_hashes.
    return \@array_of_hashes;
}

sub print_usage {
    print <<BLOCK;
Please provide the proper number of arguments:
USAGE: perl 2_parse_GenBank_file.pl [-arguments]
-i [FILE]     : input file (Required)
-o [FILE]     : output file (Required)
                
perl 2_parse_GenBank_file.pl -i NC_009673.gbk -o NC_009673.fasta
BLOCK
    exit;
}
