#!/usr/local/bin/perl
#1_determine_SNP_flanking_length.pl
#Author: Paul Stothard
#
#This script reads a file containing comma-separated SNP marker
#sequences, and determines the length of the flanking sequence given
#for each SNP. The flanking sequence length is written to a new
#comma-separated file.
#
#USAGE: perl 1_determine_SNP_flanking_length.pl [-arguments]
#-i [FILE]     : input file (Required)
#-o [FILE]     : output file (Required)
#
#perl 1_determine_SNP_flanking_length.pl -i SNP_sequences.csv -o SNP_output.csv

use warnings;
use strict;

#The Data::Dumper module provides a function called 'Dumper', which is
#useful for printing the contents of variables.
use Data::Dumper;

#The Getopt::Long module provides a function called 'GetOptions',
#which can be used to process command-line options.
use Getopt::Long;

#Create a hash to hold user-supplied options. This hash can be passed
#to subroutines, so that they can access the values.
my %options = (
    input_file  => undef,
    output_file => undef
);

#GetOptions parses command line options. 'i=s' indicates that one of
#the options is specified using '-i', and that its value is a
#string (indicated by the '=s'). The value for the '-i' option is
#stored in the %options hash, under the key 'input_file'. 'o=s'
#indicates that the other option is specified using '-o', and that its
#value is a string. The value for the '-o' option is stored in the
#%options hash, under the key 'output_file'. Other value types that
#can be specified include integer (using '=i'), real number (using
#'=f'), and flag (omit the '='). For more information on using
#GetOptions, see http://perldoc.perl.org/Getopt/Long.html
GetOptions(
    'i=s' => \$options{input_file},
    'o=s' => \$options{output_file}
);

#Check that values for the input file and output file were provided by
#the user. If an input file name isn't specified using '-i',
#$options{input_file} will not be defined, and the 'print_usage'
#subroutine will be called. The 'print_usage' subroutine prints a
#message describing how to run the script properly.
if ( !defined( $options{input_file} ) ) {
    print_usage();
}

#If an output file name isn't specified using '-o',
#$options{output_file} will not be defined, and the 'print_usage'
#subroutine will be called.
if ( !defined( $options{output_file} ) ) {
    print_usage();
}

print "Lines read from file:\n";

#The following statement calls the 'read_lines' subroutine with the
#value stored in $options{input_file}. The value returned by this call
#is stored in the array @lines. Note that the 'read_lines' subroutine
#is defined elsewhere in this script (i.e. it is not a built-in Perl
#function). Subroutines that are not part of Perl or part of an
#imported module (i.e. imported by the 'use' statement) must be
#defined in the script.
my @lines = read_lines( $options{input_file} );

#Print the contents of @lines using the Dumper function.
print Dumper(@lines);

#Call the 'add_flanking_length' subroutine. In the statement below, a
#backslash is used to create a reference to @lines. References are
#useful because they allow arrays, hashes, and complex data
#structures (an array of hashes for example) to be passed to a
#subroutine. A subroutine that is passed a reference can access or
#modify the original values stored in the variable. In this case,
#'add_flanking_length_to_lines' modifies each element in the @lines
#array. Note that 'add_flanking_length_to_lines' modifies @lines,
#rather than returning an array. This behavior differs from that of
#the 'read_lines' function--it returns an array. Generally you have to
#read the documentation of a subroutine, or write it yourself, to
#understand what it does to the values it is passed, and what it
#returns (e.g. nothing, a reference, a single value, an array).
add_flanking_length_to_lines( \@lines );
print "Lines after calling the 'add_flanking_length_to_lines'
#subroutine:\n";

#Print the contents of @lines again. Now the values in the array
#should be different, because 'add_flanking_length_to_lines' modified
#them.
print Dumper(@lines);

#Finally, call 'write_lines_to_file' with the name of the file to
#create ($options{output_file}) and a reference to @lines.
write_lines_to_file( $options{output_file}, \@lines );

#This section of the script contains several subroutines. In this
#script, most of the code is present in subroutines. This style of
#programming can allow code to be reused, and it makes for code that
#is easier to modify and debug. Typically, when writing a script you
#create a subroutine first, and then test it by passing it some
#values. Once the subroutine is working you can begin writing the next
#one, and testing it. The final code consists of several subroutines
#and a few statements in the main section of the script (i.e. not in
#subroutines).

#This subroutine reads in lines from a file, and returns them as an
#array.
sub read_lines {

    #The 'shift' statement returns the first argument passed to the
    #subroutine. In this case it is stored in $file.
    my $file = shift;
    open( my $FILE, '<', $file )
        or die("Cannot open file '$file': $!");
    my @contents = <$FILE>;
    close($FILE) or die("Cannot close file: $!");
    foreach (@contents) {
        chomp($_);
    }

    #The return value is the value of the subroutine when it is
    #called elsewhere in the script. This subroutine returns an array.
    return @contents;
}

#This subroutine accepts a reference to an array, and modifies each
#element.
sub add_flanking_length_to_lines {

    #Store the array reference in $array_ref.
    my $array_ref = shift;

    #To access an element in an array reference, use '->' between the
    #variable name and the index. In the statement below, the string
    #",total flanking sequence" is added to the first element in the
    #array.
    $array_ref->[0] = $array_ref->[0] . ",total flanking sequence";

    #To treat the array reference like an array again, surround it
    #with '@{}'. In the statement below, the 'scalar' function is
    #passed the array that is referenced by $array_ref. In the 'for'
    #loop the counting is started at '1', because the first line in
    #the input file contains column titles rather than data.
    for ( my $i = 1; $i < scalar( @{$array_ref} ); $i++ ) {

        #Add the value returned by 'get_flanking_amount' to element
        #$i. Note that 'get_flanking_amount' receives its data from
        #another subroutine ('get_cell_value'). When the below
        #statement is executed, get_cell_value is called first,
        #followed by 'get_flanking_amount.
        $array_ref->[$i] = $array_ref->[$i] . ','
            . get_flanking_amount( get_cell_value( 1, $array_ref->[$i] ) );
    }
}

#This subroutine accepts a column number and a row of data. It splits
#the row into an array of values, and returns the value stored in
#element 'column number'.
sub get_cell_value {
    my $column_index = shift;
    my $row_contents = shift;
    my @cells        = split( /,/, $row_contents );
    return $cells[$column_index];
}

#This subroutine uses a regular expression to determine the amount of
#sequence on either side of a variable base (indicated by square
#brackets in this case).
sub get_flanking_amount {
    my $value   = shift;
    my $amount = undef;

    #The regular expression below looks for text like the following:
    #ACTGATGGCACACAATC[T/C]CTTGGGGACCTGTTA
    #and matches the sequence before and after the square brackets.
    #The 'x' at the end of the regular expression causes whitespace in
    #the regular expression statement to be ignored. When whitespace
    #is ignored, the regular expression can be written over multiple
    #lines, and comments can be added for important sections.
    if ($value =~ m/          
                  (        #Capture the upstream sequence in $1.
                    [^\[]+ #One or more non-bracket characters.
                  )        #End of $1 capture.
                  \[       #Opening square bracket
                  \S       #First variable base
                  \/       #Forward slash
                  \S       #Second variable base
                  \]       #Closing square bracket
                  (        #Capture the downstream sequence in $2.
                    \S+    #One or more non-whitespace characters.
                  )        #End of $2 capture.
                  /x
        )
    {
        #The matches that are 'remembered' (because of the parentheses
        #in the regular expression) can be accessed as $1, $2.
        $amount = length($1) + length($2);
    }
    else {
        #If the value doesn't match the expected pattern, terminate
        #execution.
        die("The following value could not be parsed: $value");
    }
    return $amount;
}

#This subroutine accepts a file name and an array reference. It then
#writes the contents of the array to the specified file.
sub write_lines_to_file {
    my $file      = shift;
    my $array_ref = shift;
    open( my $FILE, '>', $file )
        or die("Cannot open file '$file': $!\n");

    #The array reference is treated like an array again, because it is
    #dereferenced.
    foreach ( @{$array_ref} ) {
        print $FILE $_ . "\n";
    }
    close($FILE) or die("Cannot close file: $!");
}

#The print_usage subroutine prints a message describing how to use
#this script. After printing the message the program terminates,
#because of the 'exit' statement.
sub print_usage {

    #The following statement prints the text contained between
    #'<<BLOCK;' and 'BLOCK'.
    print <<BLOCK;
Please provide the proper number of arguments:
USAGE: perl 1_determine_SNP_flanking_length.pl [-arguments]
-i [FILE]     : input file (Required)
-o [FILE]     : output file (Required)
                
perl 1_determine_SNP_flanking_length.pl -i SNP_sequences.csv -o SNP_output.csv
BLOCK
    exit;
}
