#!/usr/local/bin/perl
#3_local_blast.pl
#Author: Paul Stothard
#
#This script formats a sequence database using formatdb, and then
#submits a series of sequences to the blastall program. The results of
#the searches are written to separate files in a specified directory.
#
#USAGE: perl 3_local_blast.pl [-arguments]
#-i [FILE]     : input file (Required)
#-d [FILE]     : database file (Required)
#-o [FILE]     : output directory (Required)
#
#perl 3_local_blast.pl -i query_sequences.fasta -d DNA_database.fasta -o blast_results

use warnings;
use strict;

use Data::Dumper;
use Getopt::Long;

#The File::Temp module can be used to safely create temporary files
#that are automatically removed after a script has run. Creating
#temporary files without this module can be risky. For example, if
#multiple copies of a script are run simultaneously there may be cases
#where temporary files are overwritten or the wrong temporary file is
#accessed. For more information on this module, see:
#http://search.cpan.org/~tjenness/File-Temp-0.18/Temp.pm
use File::Temp;

#Store user-supplied options and configuration information (the paths
#to the formatdb program and the blastall program) in a hash.
my %options = (
    input_file              => undef,
    database_file           => undef,
    output_directory        => undef,
    formatdb_path           => '/usr/local/blast/bin/formatdb',
    blastall_path           => '/usr/local/blast/bin/blastall',
    is_protein_database     => 'F',
    blast_program           => 'blastn',
    formatted_database_file => undef
);

GetOptions(
    'i=s' => \$options{input_file},
    'd=s' => \$options{database_file},
    'o=s' => \$options{output_directory}
);

if ( !defined( $options{input_file} ) ) {
    print_usage();
}
if ( !defined( $options{database_file} ) ) {
    print_usage();
}
if ( !defined( $options{output_directory} ) ) {
    print_usage();
}

#Check to see if the output directory exists, using the '-d' file test
#operator. If the directory doesn't exist, create it using the mkdir
#function. The second argument to mkdir (0775) is used to specify the
#file permissions associated with the directory.
if ( !( -d $options{output_directory} ) ) {
    mkdir( $options{output_directory}, 0775 )
        or die("Cannot create directory :$!");
}

#Call a subroutine that formats the BLAST database.
format_blast_database( \%options );

#Read sequences from the input file.
open( my $INFILE, '<', $options{input_file} ) or die("Cannot open file : $!");
local $/ = ">";
my @blast_results = ();
while ( my $data = <$INFILE> ) {
    if ( $data eq '>' ) {
        next;
    }

    #Convert the FASTA sequence to a hash reference.
    my $sequence_record = parse_fasta( '>' . $data );

    #Obtain the BLAST results for the sequence.
    my $blast_results
        = get_local_blast_results( \%options, $sequence_record );

    #Write the BLAST results to separate files, named after the query.
    if ( defined($blast_results) ) {
        my $results_file
            = "$options{output_directory}"
            . "/$sequence_record->{title}"
            . ".blast";
        if ( -e $results_file ) {
            print "Warning: overwriting previous results written to $results_file\n";
        }

        open( my $OUTFILE, '>', $results_file )
            or die("Cannot open file : $!");
        print $OUTFILE $blast_results . "\n";
        close($OUTFILE) or die("Cannot close file: $!");
    }
}

close($INFILE) or die("Cannot close file: $!");

#Call the formatdb program, to format the sequence database.
sub format_blast_database {
    my $options = shift;

    #Obtain a safe temporary filename.
    my $temp_file_handle = new File::Temp();
    $options{formatted_database_file} = $temp_file_handle->filename;

    #Create the command for calling the formatdb program.
    my $format_db_command
        = "$options{formatdb_path}"
        . " -i $options{database_file}"
        . " -n $options{formatted_database_file}"
        . " -p $options{is_protein_database}";

    #Use the system command to run the command stored in
    #$format_db_command.
    my $exit_value = system($format_db_command);

    #$exit_value will be non-zero if a problem is encountered.
    if ( $exit_value != 0 ) {
        print "$format_db_command\n";
        print "failed to execute : $!\n";

        #Exit with a value of '1', to indicate that an error was encountered.
        exit(1);
    }
}

#Convert a FASTA sequence to a hash containing a title and a sequence.
sub parse_fasta {
    my $data         = shift;
    my %fasta_record = (
        title    => undef,
        sequence => undef
    );

    if ( $data =~ m/^\s*>([^\n\cM]+)/ ) {
        $fasta_record{title} = $1;
    }

    $data =~ s/^[^\n\cM]+//;
    $data =~ s/[^A-Z]//ig;

    $fasta_record{sequence} = $data;

    return \%fasta_record;
}

#Call the blastall program, to perform a BLAST search.
sub get_local_blast_results {
    my $options         = shift;
    my $sequence_record = shift;

    my $temp_file_handle = new File::Temp();
    my $temp_file_name   = $temp_file_handle->filename;

    print $temp_file_handle
        ">$sequence_record->{title}\n$sequence_record->{sequence}\n";
    close($temp_file_handle) or die("Cannot close file: $!");

    my $blast_command
        = "$options{blastall_path}"
        . " -p $options{blast_program}"
        . " -d $options{formatted_database_file}"
        . " -i $temp_file_name";

    #Backticks are used in the following statement to execute the
    #command stored in $blast_command. The output generated by the
    #BLAST search is stored in $blast_output.
    my $blast_output = `$blast_command`;
    return $blast_output;
}

sub print_usage {
    print <<BLOCK;
Please provide the proper number of arguments:
USAGE: perl 3_local_blast.pl [-arguments]
-i [FILE]     : input file (Required)
-d [FILE]     : database file (Required)
-o [FILE]     : output directory (Required)

perl 3_local_blast.pl -i query_sequences.fasta -d DNA_database.fasta -o blast_results
BLOCK
    exit;
}
