#!/usr/local/bin/perl
#4_read_blast_results.pl
#Author: Paul Stothard
#
#This script reads a tab-delimited file containing BLAST results and
#builds a complex data structure. The script then sorts the hits and
#writes them to a new file.
#
#USAGE: perl 4_read_blast_results.pl [-arguments]
#-i [FILE]     : input file (Required)
#-o [FILE]     : output file (Required)
#
#perl 4_read_blast_results.pl -i blast_results.tab -o sorted_blast_results.tab
use warnings;
use strict;

use Data::Dumper;
use Getopt::Long;

my %options = (
    input_file  => undef,
    output_file => undef
);

GetOptions(
    'i=s' => \$options{input_file},
    'o=s' => \$options{output_file}
);

if ( !defined( $options{input_file} ) ) {
    print_usage();
}
if ( !defined( $options{output_file} ) ) {
    print_usage();
}

#Convert the input file to an array of lines.
my $lines = read_lines( $options{input_file} );

#Get rid of the first element in @lines, as it contains column titles.
shift( @{$lines} );

#Create a hash where each key is a query name, and each value is an
#array of hashes. Store this data structure in $hash_of_hits.
my $hash_of_hits = get_blast_hits_hash($lines);

#Get the keys from the hash reference by dereferencing $hash_of_hits
#back into a hash (using %{$hash_of_hits}).
my @keys = keys( %{$hash_of_hits} );

#Sort the keys alphabetically. The code block { lc($a) cmp lc($b) } is
#used to determine how the elements in @keys are sorted. Sorting code
#blocks return an integer less than, equal to, or greater than 0. In
#the code block in this example, the lowercase versions of the keys
#are compared using the 'cmp' operator, which returns 1 if lc($a) is
#greater than lc($b) (based on ASCII values of each character in the
#strings), -1 if lc($a) is less than lc($b), and 0 if they are the
#same. For more information on sorting arrays, see
#http://perldoc.perl.org/functions/sort.html
@keys = sort { lc($a) cmp lc($b) } @keys;

open( my $OUTFILE, '>', $options{output_file} )
    or die("Cannot open file '$options{output_file}': $!\n");

#Print column titles.
print $OUTFILE "query_id\tmatch_id\talignment_length\tevalue\n";

#Write the BLAST results out to a file. Because the keys have been
#sorted, the results will be sorted by query id.
foreach (@keys) {
    my $query_id = $_;

    #$hash_of_hits is a reference to a hash, so to access values, use
    #'->' between the variable name and the key. In this case the
    #value is an array reference.
    my $array_ref = $hash_of_hits->{$query_id};

    #Sort the hits by evalue, followed by alignment length. To do
    #this, first use the map function to build a list containing the
    #values to be used in sorting. The following statement takes each
    #element from the array @{$array_ref}, and builds a new array
    #consisting of three elements: the evalue, the alignment length,
    #and the element itself. Each of the new arrays generated by map
    #is added to @precomputed, as an array reference. This sorting
    #strategy is especially useful when sorting is performed using
    #calculated values, since the calculations only need to be
    #performed once for each element of the array. In this example
    #calculated values are not used--just stored values.
    my @precomputed
        = map { [ $_->{evalue}, $_->{alignment_length}, $_ ] } @{$array_ref};

    #To sort @precomputed, use the sort function. The elements in
    #@precomputed are array references, so '->' is used to access the
    #values. The sort portion of this statement works as follows: the
    #evalues of hits are compared first ($a->[0] <=> $b->[0]). If the
    #evalues are the same, the comparison returns a value of '0'. The
    #'||' operator causes the second comparison ($a->[1] <=> $b->[1])
    #to be performed when the first returns '0'. The second comparison
    #compares alignment lengths.
    my @ordered_precomputed
        = sort { $a->[0] <=> $b->[0] || $a->[1] <=> $b->[1] } @precomputed;

    #To get the final sorted array, obtain the third element from the
    #array references in @ordered_precomputed. Recall that the third
    #element is the hash reference containing the BLAST hit
    #information.
    my @ordered = map { $_->[2] } @ordered_precomputed;

    #Write out some information about each hit to file.
    foreach (@ordered) {
        my $hash_ref = $_;
        print $OUTFILE "$query_id\t"
            . "$hash_ref->{match_id}\t"
            . "$hash_ref->{alignment_length}\t"
            . "$hash_ref->{evalue}\n";
    }
}
close($OUTFILE) or die("Cannot close file: $!");

#Read lines from an input file.
sub read_lines {
    my $file     = shift;
    my @contents = ();
    open( my $FILE, '<', $file ) or die("Cannot open file '$file': $!\n");
    while ( my $line = <$FILE> ) {

        #Skip lines that start with '#'.
        if ( $line =~ m/^\#/ ) {
            next;
        }
        chomp($line);
        push( @contents, $line );
    }
    close($FILE) or die("Cannot close file: $!");
    return \@contents;
}

#Build a hash of arrays of hashes.
sub get_blast_hits_hash {
    my $blast_results = shift;

    #This hash will contain query names as the keys.
    my %hash_of_hits = ();
    foreach ( @{$blast_results} ) {
        my @fields = split( /\t/, $_ );
        my $query_id = $fields[0];

        #Some rows in the sample data are missing fields. Skip these
        #rows.
        if ( scalar(@fields) < 13 ) {
            next;
        }

        #If the query name isn't in the hash, add it to the hash and
        #store a reference to a new array.
        if ( !( exists( $hash_of_hits{$query_id} ) ) ) {
            my @array = ();
            $hash_of_hits{$query_id} = \@array;
        }

        #Create a hash to store information about the BLAST hit.
        #The fields in the BLAST file are:
        #query_id, match_id, match_description, %_identity,
        #alignment_length, mismatches, gap_openings, q_start, q_end,
        #s_start, s_end, evalue, bit_score
        my %hit_info = ();
        $hit_info{match_id}         = $fields[1];
        $hit_info{identity}         = $fields[3];
        $hit_info{alignment_length} = $fields[4];
        $hit_info{evalue}           = $fields[11];
        $hit_info{bit_score}        = $fields[12];

        #Add a reference to %hit_info to the array for this query.
        push( @{ $hash_of_hits{$query_id} }, \%hit_info );
    }
    return \%hash_of_hits;
}

sub print_usage {
    print <<BLOCK;
Please provide the proper number of arguments:
USAGE: perl 4_read_blast_results.pl [-arguments]
-i [FILE]     : input file (Required)
-o [FILE]     : output file (Required)

perl 4_read_blast_results.pl -i blast_results.tab -o sorted_blast_results.tab 
BLOCK
    exit;
}
